//
// jja: swiss army knife for chess file formats
// src/merge.rs: PolyGlot Book Merging
//
// Copyright (c) 2023 Ali Polatel <alip@chesswob.org>
//
// SPDX-License-Identifier: GPL-3.0-or-later

use crate::polyglot::BookEntry;

/// Represents various strategies for merging two Polyglot opening books.
///
/// When merging two opening books, different strategies can be used to determine
/// how to combine the weights of the same moves in different books.
#[derive(Copy, Clone, Debug, PartialEq)]
pub enum MergeStrategy {
    /// Calculates the average weight of the moves in both books.
    AvgWeight,
    /// Takes the maximum weight of the moves in both books.
    MaxWeight,
    /// Takes the minimum weight of the moves in both books.
    MinWeight,
    /// Prioritizes the moves of the first book, ignoring moves from the second book.
    Ours,
    /// Calculates the sum of the weights of the moves in both books.
    SumWeight,
    /// Calculates the weighted average weight of the moves in both books, considering the given
    /// weights for each book.
    ///
    /// The tuple contains the weights associated with the first and second books, respectively.
    WeightedAverageWeight(f64, f64),
    // Add more strategies here
}

/// Calculates the average of the weights of two book entries.
///
/// # Arguments
///
/// * `entry1` - The first book entry.
/// * `entry2` - The second book entry.
///
/// # Returns
///
/// * A `u16` representing the average weight of the two book entries.
///   In case of overflow, it returns `u16::MAX`.
pub fn average_weight(entry1: &BookEntry, entry2: &BookEntry) -> u16 {
    let entry1_weight: u32 = entry1.weight.into();
    let entry2_weight: u32 = entry2.weight.into();
    let sum = entry1_weight.saturating_add(entry2_weight);
    (sum / 2).try_into().unwrap_or(u16::MAX)
}

/// Calculates the weighted average of the weights of two book entries,
/// considering the given weights for each book.
///
/// # Arguments
///
/// * `entry1` - The first book entry.
/// * `entry2` - The second book entry.
/// * `book1_weight` - The weight associated with the first book.
/// * `book2_weight` - The weight associated with the second book.
///
/// # Returns
///
/// * A `u16` representing the weighted average weight of the two book entries,
///   taking into account the weights of each book.
///   In case of overflow, it returns `u16::MAX`.
pub fn weighted_average_weight(
    entry1: &BookEntry,
    entry2: &BookEntry,
    book1_weight: f64,
    book2_weight: f64,
) -> u16 {
    let total_weight = book1_weight + book2_weight;
    let normalized_book1_weight = book1_weight / total_weight;
    let normalized_book2_weight = book2_weight / total_weight;

    let entry1_weight: f64 = entry1.weight.into();
    let entry2_weight: f64 = entry2.weight.into();

    let weighted_sum =
        entry1_weight * normalized_book1_weight + entry2_weight * normalized_book2_weight;
    weighted_sum.min(f64::from(u16::MAX)).round() as u16
}
